;LCD-Routinen fr WD-C2704M - Text-Punktmatrix-LCD
;mit 2 Controllern HD44780. 4 Zeilen je 27 Zeichen

;Version mit frei definierbaren Steuerleitungen.

;Diese Routinensammlung baut auf Beispielen von Andreas Schwarz
;und T.S. auf und ist von ...HanneS... zusammengestellt worden.
; hannes.lux@gmx.de / www.hanneslux.de

;============================================================================
;Diese Sammlung enthlt folgende Routinen:
;----------------------------------------------------------------------------
;  lcd_init             ;initialisiert und lscht beide Controller des LCD
;----------------------------------------------------------------------------
;  lcd_clear            ;lscht das LCD
;----------------------------------------------------------------------------
;  lcd_command          ;bergibt das Byte im Register "tmp" als Kommando
                        ;an das LCD
;----------------------------------------------------------------------------
;  lcd_data             ;bergibt das Byte (ASCII-Zeichen) im Register "tmp" 
                        ;zur Textausgabe. Dabei sind folgende Steuerzeichen
                        ;mglich:
                        ;1:  Blinken ausschalten
                        ;2:  akt. Ausgabe-Position als Blinkanfang setzen
                        ;3:  akt. Ausgabe-Position als Blinkende setzen
                        ;4:  Bildschirm lschen
                        ;8:  Cursor home (ohne Lschen)
                        ;9:  Locate, Position folgt im nchsten Byte
                        ;13: CR, Zeilenwechsel, Rest der Zeile wird gelscht
;----------------------------------------------------------------------------
;  locate			;Macro zum Einstellen der Ausgabeposition.
                        ;Aufruf: locate Zeile,Spalte
                        ;Zeile: 0-3, Spalte: 0-39
;============================================================================
;Die Datei "LCDprint.inc" stellt einige Ausgaberoutinen zur Verfgung, die
;die Ausgabe von Texten und Zahlen erleichtert. Zusammen mit "locate" wird
;die Ausgabe von Mentexten, Zahlen usw. hnlich einfach wie in BASIC.
;============================================================================

;Zur Steuerung der LCD-Zugriffe wird ein oberes Register namens "VFLAGS"
;fr einige Booleans bentigt. Es muss im Hauptprogramm deklariert werden.
;Die eizelnen Bits werden hier deklariert: 
.equ zeiwe=1                ;Flag fr Zeilenwechsel
.equ update=2               ;Flag fr Update LCD
.equ upsync=3               ;Synchronisationsflag vom Timer-Interrupt
.equ neu=4                  ;neues Byte eingetroffen
.equ loc=5                  ;Flag fr Locate-Modus
.equ lcdcontroller1=6       ;Flag fr Controller 1
.equ lcdcontroller2=7       ;Flag fr Controller 2
;Die bergabe der Daten erfolgt ber das Register "wl", das im Hauptprogramm
;zu deklarieren ist. Weiterhin muss das Register "wh" deklariert sein.

;============================================================================
;Anschlussbelegung:     ;Die 4 Datenleitungen sollten in einem Nibble liegen.
                        ;Dies kann ein unteres oder oberes Nibble sein.
                        ;Die Auswahl geschieht mit "lcddatmsk"
                        ;Die Steuerleitungen des LCD knnen ber alle
                        ;verfgbaren Ports frei gewhlt werden, was diesen
                        ;Treiber sehr flexibel macht.
;----------------------------------------------------------------------------
.equ lcdport=portd      ;Port, an dem die 4 Datenleitungen des LCD
                        ;angeschlossen sind.
.equ lcddatmsk=$f0      ;Maske auf LCD-Datenpins.
                        ;$0f - LCD ist an Bit 0...3 des Ports angeschlossen.
                        ;$f0 - LCD ist an Bit 4...7 des Ports angeschlossen.
;----------------------------------------------------------------------------
.equ rsport=porta       ;Port, an dem RS angeschlossen ist
.equ rs=7               ;Pin, an dem RS angeschlossen ist
;----------------------------------------------------------------------------
.equ enable1port=portd  ;Port Anschluss Enable des ersten Controllers.
.equ enable1=2          ;Pin Anschluss Enable des ersten Controllers.
;----------------------------------------------------------------------------
.equ enable2port=portd  ;Port Anschluss Enable des zweiten Controllers.
.equ enable2=3          ;Pin Anschluss Enable des zweiten Controllers.
;----------------------------------------------------------------------------
;.equ rwport=portd       ;Port, an dem R/W angeschlossen ist
;.equ rw=2               ;Pin, an dem R/W angeschlossen ist
;============================================================================


;=============================================================================
;  locate Zeile,Spalte  ;Setzt die Ausgabeposition im VLCD auf
                        ;Zeile (0...3) und Spalte (0...26).
;-----------------------------------------------------------------------------
.macro locate ;Zeile (0...3), Spalte (0...27)
                        ;Positionierung der Ausgabeposition
 push wl                    ;Register sichern
 ldi wl,((@0 & 3)*spn)+(@1 & 63)+low(vlcd)  ;Position im VLCD errechnen
 sts wpos,wl                ;und einstellen
 pop wl                     ;Register wiederherstellen
.endmacro


;=============================================================================
;  blink Zeile,Spalte,Anzahl  ;Markiert die Blinkposition im VLCD auf
                        ;Zeile (0...3), Spalte (0...26) und Anzahl blinkender
                        ;Zeichen.
;-----------------------------------------------------------------------------
.macro blink ;Zeile (0...3), Spalte (0...27), Anzahl Blinkzeichen
                        ;Positionierung der Blinkposition
 push wl                    ;Register sichern
 ldi wl,((@0 & 3)*spn)+(@1 & 63)+low(vlcd)  ;Startposition im VLCD errechnen
 sts blinka,wl              ;und einstellen
 subi wl,-@2                ;Endposition im VLCD
 sts blinke,wl              ;einstellen
 pop wl                     ;Register wiederherstellen
.endmacro


;============================================================================
; Aufrufbare Routine, initialisiert und lscht das LCD
;----------------------------------------------------------------------------
lcd_init:               ;LCD-Initialisierung
 ldi zl,low(vlcd)           ;Z-Pointer
 ldi zh,high(vlcd)          ;auf VLCD
 sts rpos,zl                ;VLCD auf Anfang des
 sts wpos,zl                ;Bildschirmspeichers setzen
 sts blinka,zl              ;Blinkpointer
 sts blinke,zl              ;auch
 ldi tmp,' '                ;Leerzeichen
lcd_init1:
 st z+,tmp                  ;ins VLCD
 cpi zl,low(vlcd+zln*spn)   ;Ende VLCD erreicht?
 brne lcd_init1             ;nein, nochmal...

 sbr vflags,(1<<lcdcontroller1)|(1<<lcdcontroller2)
                            ;erstmal beide LCD-Controller aktivieren

 in wl,lcdport-1            ;Datenrichtung LCD-Datenport holen
 sbr wl,lcddatmsk           ;benutzte Datenpins als Ausgang
 out lcdport-1,wl           ;definieren
 sbi rsport-1,rs            ;RS-Leitung als Ausgang
 sbi enable1port-1,enable1  ;Enable1-Leitung als Ausgang
 sbi enable2port-1,enable2  ;Enable2-Leitung als Ausgang
; sbi rwport-1,rw            ;Read/Write-Leitung als Ausgang

 push xl                    ;Inhalt sichern
; ldi xl,50                  ;>15ms warten, nachdem Vcc > 4,5V ist
 ldi xl,20                  ;>15ms warten, nachdem Vcc > 4,5V ist
powerupwait:                ;hier 250ms warten, dafr keine Spannungsprfung
 rcall delay5ms
 dec xl
 brne powerupwait
 pop xl                     ;alten Inhalt zurck holen

; cbi rwport,rw              ;Steuerleitungen auf
 cbi rsport,rs              ;L-Pegel legen
;Startsequenz
 in tmp,lcdport              ;ganzen Port einlesen
 cbr tmp,lcddatmsk           ;Platz fr Datenbits lschen
 ldi wh,0b00110011          ;bentigte Datenbits in beide Nibbles
 andi wh,lcddatmsk          ;nur das genutzte Nibble
 or tmp,wh                   ;bernehmen
 out lcdport,wl             ;und wieder ausgeben
;dreimal ausgeben
 rcall lcd_impuls           ;1. Ausgabe (von 3)
 rcall delay5ms             ;>4,1ms warten
 rcall lcd_impuls           ;2
 rcall delay5ms             ;>0,1ms warten
 rcall lcd_impuls           ;und 3!
 rcall delay5ms             ;Ab hier BUSY prfen
;4-Bit-Modus aktivieren
 in tmp,lcdport              ;ganzen Port einlesen
 cbr tmp,lcddatmsk           ;Platz fr Datenbits lschen
 ldi wh,0b00100010          ;bentigte Datenbits in beide Nibbles
 andi wh,lcddatmsk          ;nur das genutzte Nibble
 or tmp,wh                   ;bernehmen
 out lcdport,wl             ;und wieder ausgeben
 rcall lcd_impuls
 rcall delay5ms
;ab jetzt 4-Bit-Zugriff mit Busy-Abfrage
 ldi tmp,0b00101000          ;4bit-Modus einstellen
 rcall lcd_command
 rcall delay5ms
 ldi tmp,0b00001100          ;Display ein, Cursor aus
 rcall lcd_command
 rcall delay5ms
 ldi tmp,0b00000100          ;Entry-Mode
 rcall lcd_command
 rcall delay5ms
                            ;(Code geht weiter...)

;============================================================================
; Aufrufbare Routine, lscht das LCD
;----------------------------------------------------------------------------
lcd_clear:  ;Sendet den Befehl zur Lschung des Displays
 sbr vflags,(1<<lcdcontroller1)|(1<<lcdcontroller2) ;beide Controller
 ldi tmp,0b00000001          ;Display lschen
 rcall lcd_command
 cbr vflags,1<<lcdcontroller2;nur ersten Controller aktiviert lassen
 rcall delay5ms
 ret


;============================================================================
; Aufrufbare Routine, sendet den Befehl in wl an das LCD
;----------------------------------------------------------------------------
lcd_command:            ;sendet den Befehl in wl an das LCD
 push wh                    ;fr Nibble-Wahl
 ldi wh,lcddatmsk           ;Referenz fr Nibblewahl
 push tmp                   ;Commandobyte sichern
 sbrc wh,0                  ;wenn LCD am unteren Nibble angeschlossen ist
 swap tmp                   ;dann oberes Nibble nach unten
 cbi rsport,rs              ;RS lschen
 rcall lcd_nibble           ;erstes Nibble ausgeben...
 pop tmp                    ;Kopie vom Commandobyte holen
 sbrc wh,4                  ;wenn LCD am oberen Nibble angeschlossen ist
 swap tmp                   ;dann unteres Nibble nach oben
 pop wh                     ;wieder frei
 cbi rsport,rs              ;RS lschen
 rjmp lcd_nibble            ;zweites Nibble ausgeben und zurck...


;============================================================================
; Aufrufbare Sub-Routine, legt das Zeichen in wl in das VLCD
;----------------------------------------------------------------------------
lcd_data:               ;UP, schreibt Zeichen in wl in virtuellen Bildschirm
 push zh
 push zl
 ldi zh,high(vlcd)          ;Z-Pointer auf
 lds zl,wpos                ;Ausgabeposition
 sbrs vflags,loc            ;Locate-Modus aktiv? - ja...
 rjmp lcd_nixlocate         ;nein...
 add zl,tmp                 ;Offset addieren
 cbr vflags,1<<loc          ;Statusflag Locate deaktivieren
 rjmp lcd_data1             ;Position speichern und Ende...
lcd_nixlocate:          ;es war kein Locate
 cpi tmp,32                 ;Zeichen unter Chr$(32)?
 brmi lcd_stz               ;ja, als Steuerzeichen behandeln...
 st z+,tmp                  ;nein, Zeichen ins VLCD schreiben
 cpi zl,low(vlcd+zln*spn)   ;letztes Zeichen erreicht?
 brne lcd_data1             ;nein...
lcd_data0:
 ldi zl,vlcd                ;ja, Pointer auf Anfang des Bildschirmspeichers
lcd_data1:
 sts wpos,zl                ;Pointer wieder sichern
 rjmp lcd_data_end          ;fertig...
lcd_stz:                ;in WL ist ein Steuerzeichen
 cpi tmp,13                 ;Zeilenumbruch?
 brne lcd_stz1              ;nein...
 ldi tmp,' '                ;Leerzeichen
lcd_stz0:               ;
 st z+,tmp                  ;Leerzeichen ins VLCD schreiben
 cpi zl,low(vlcd+spn*1)     ;letztes Zeichen der Zeile erreicht?
 breq lcd_data1             ;ja, Pointer sichern und fertig...
 cpi zl,low(vlcd+spn*2)     ;letztes Zeichen der Zeile erreicht?
 breq lcd_data1             ;ja, Pointer sichern und fertig...
 cpi zl,low(vlcd+spn*3)     ;letztes Zeichen der Zeile erreicht?
 breq lcd_data1             ;ja, Pointer sichern und fertig...
 cpi zl,low(vlcd+spn*4)     ;letztes Zeichen der Zeile erreicht?
 breq lcd_data0             ;ja, Pointer korregieren, sichern und fertig...
 rjmp lcd_stz0              ;nochmal...
lcd_stz1:               ;kein Zeilenumbruch
 cpi tmp,4                  ;CLS?
 brne lcd_stz_cls           ;nein...
 ldi zl,low(vlcd)           ;Pointer auf
 ldi zh,high(vlcd)          ;Beginn Bildschirmspeicher
 sts wpos,zl                ;Ausgabeposition,
 sts blinka,zl              ;Blinkanfang,
 sts blinke,zl              ;und Blinkende sichern
 ldi tmp,32                 ;Leerzeichen
lcd_stz_cls0:           ;Loop CLS
 st z+,tmp                  ;Leerzeichen in Bildschirmspeicher schreiben
 cpi zl,low(vlcd+spn*zln)   ;letztes Zeichen der letzten Zeile erreicht?
 brne lcd_stz_cls0          ;nein, nochmal...
 rjmp lcd_data_end          ;fertig...
lcd_stz_cls:            ;Ende CLS
 cpi tmp,1                  ;Blinken aus?
 brne lcd_stz_bls           ;nein...
 ldi zl,low(vlcd)           ;ja, Beginn Bildschirmspeicher
 sts blinka,zl              ;als Blinkanfang
 sts blinke,zl              ;und Blinkende sichern
lcd_stz_bls:            ;Ende Blinken aus
 cpi tmp,2                  ;Blinkanfang?
 brne lcd_stz_bla           ;nein...
 lds zl,wpos                ;ja, aktuelle Ausgabeposition
 dec zl                     ;
 sts blinka,zl              ;als Blinkanfang sichern
lcd_stz_bla:            ;Ende Blinkanfang setzen
 cpi tmp,3                  ;Blinkende?
 brne lcd_stz_ble           ;nein...
 lds zl,wpos                ;ja, aktuelle Ausgabeposition
 dec zl                     ;
 sts blinke,zl              ;als Blinkende sichern
lcd_stz_ble:            ;Ende Blinkende setzen
 cpi tmp,9                  ;Locate?
 brne lcd_stz_loc           ;nein...
 sbr vflags,1<<loc          ;ja, Status-Flag fr Locate setzen
 rjmp lcd_data0             ;Position 0 speichern und Ende...
lcd_stz_loc:            ;Ende Locate
 cpi tmp,8                  ;Cursor Home?
 brne lcd_stz_home          ;nein...
 rjmp lcd_data0             ;Position 0 speichern und Ende...
lcd_stz_home:           ;Ende Cursor home

lcd_stz_stz2:           ;noch keine weiteren Kommandos implementiert
lcd_data_end:
 sbr vflags,1<<update       ;Update anmelden
 pop zl
 pop zh
 ret


;============================================================================
; Job der Mainloop, sendet das Zeichen in wl an das LCD
;----------------------------------------------------------------------------
lcd_update:             ;UP, sendet ein Zeichen aus dem VLCD an das LCD
 cbr vflags,1<<upsync       ;Jobflag lschen
 ldi zh,high(vlcd)          ;Z-Pointer auf
 lds zl,rpos                ;Ausgabeposition
 sbrc vflags,zeiwe          ;Zeilenwechsel gewesen? - nein, jetzt prfen...
 rjmp lcd_u_nixzw           ;ja, jetzt nicht prfen...

lcd_upd1:
 cpi zl,vlcd+spn            ;Ende erste Zeile?
 brne lcd_upd2              ;nein...
 sbr vflags,1<<zeiwe        ;Zeilenwechsel merken
 ldi tmp,128+64             ;set DD-RAM-Adress auf Zeile 2 (1..2)
 rjmp lcd_command           ;

lcd_upd2:
 cpi zl,vlcd+spn*2          ;Ende zweite Zeile?
 brne lcd_upd3              ;nein...
 sbr vflags,1<<zeiwe        ;Zeilenwechsel merken
 ldi tmp,128+0              ;set DD-RAM-Adress auf Zeile 1 (1..2)
 cbr vflags,1<<lcdcontroller1   ;Controller1 (1..2) aus,
 sbr vflags,1<<lcdcontroller2   ;Controller2 (3..4) ein
 rjmp lcd_command           ;

lcd_upd3:
 cpi zl,vlcd+spn*3          ;Ende dritte Zeile?
 brne lcd_upd4              ;nein...
 sbr vflags,1<<zeiwe        ;Zeilenwechsel merken
 ldi tmp,128+64             ;set DD-RAM-Adress auf Zeile 2 (1..2)
 rjmp lcd_command           ;

lcd_upd4:
 cpi zl,low(vlcd+spn*4)     ;Ende vierten Zeile?
 brne lcd_upd0              ;nein...
 inc bliz                   ;Blinkzhler erhhen
 sbr vflags,1<<zeiwe        ;Zeilenwechsel merken
 ldi tmp,128+0              ;set DD-RAM-Adress auf Zeile 1 (1..4)
 cbr vflags,1<<lcdcontroller2   ;Controller2 (3..4) aus,
 sbr vflags,1<<lcdcontroller1   ;Controller1 (1..2) ein
 ldi zl,low(vlcd)           ;von vorn beginnen
 sts rpos,zl                ;und Pointer wieder sichern
 rjmp lcd_command           ;

lcd_u_nixzw:            ;es hat beim letzten mal einen Zeilenwechsel gegeben
 cbr vflags,1<<zeiwe        ;Flag vom letzten Zeilenwechsel lschen
 cpi zl,vlcd                ;Position 0?
 brne lcd_upd0              ;nein...
 sbrc vflags,update         ;sind Zeichen im VLCD verndert worden? - nein...
 rjmp lcd_upd0              ;ja, eine Runde updaten...
 ret                        ;nein, fertig...

lcd_upd0:               ;es hat jetzt kein Zeilenwechsel stattgefunden
 ld tmp,z+                  ;Zeichen holen
 sts rpos,zl                ;Pointer sichern
 sbrs bliz,2                ;nur wenn Blinkphase aktiv ist
 rjmp lcd_byte              ;jetzt also nicht...
 dec zl                     ;alten Pointerwert wiederherstellen
 lds zh,blinka              ;Position Blinkanfang holen
 cp zh,zl                   ;Pointer noch unter Blinkanfang?
 brcc lcd_byte              ;ja, nicht blinken...
 lds zh,blinke              ;Position Blinkende holen
 cp zh,zl                   ;Pointer noch unter Blinkende?
 brcs lcd_byte              ;nein, nicht blinken...
 ldi tmp,' '                ;ja, Leerzeichen ausgeben
lcd_byte:
 push wh                    ;fr Nibble-Wahl
 ldi wh,lcddatmsk           ;Referenz fr Nibblewahl
 push tmp                   ;Datenbyte sichern
 sbrc wh,0                  ;wenn LCD am unteren Nibble angeschlossen ist
 swap tmp                   ;dann oberes Nibble nach unten
 sbi rsport,rs              ;RS setzen
 rcall lcd_nibble           ;erstes Nibble ausgeben...
 pop tmp                    ;Kopie vom Datenbyte holen
 sbrc wh,4                  ;wenn LCD am oberen Nibble angeschlossen ist
 swap tmp                   ;dann unteres Nibble nach oben
 pop wh                     ;wieder frei
 sbi rsport,rs              ;RS setzen
lcd_nibble:
 andi tmp,lcddatmsk         ;Daten auf vom LCD genutztes Nibble maskieren
 push wh                    ;benutztes Register sichern
 in wh,lcdport              ;ganzen Port einlesen
 cbr wh,lcddatmsk           ;nur LCD-Datenbits lschen
 or wh,tmp                  ;neue Datenbits einfgen
 out lcdport,wh             ;ausgeben
 pop wh                     ;alten Wert in Register wiederherstellen
lcd_impuls:                 ;erzeugt den Enable-Puls
 sbrc vflags,lcdcontroller1  ;Controller 1?
 sbi enable1port,enable1    ;ja, Enable1 H
 sbrc vflags,lcdcontroller2  ;Controller 2?
 sbi enable2port,enable2    ;ja, Enable2 H
 rcall lcd_impulsR          ;7 Takte warten
 rcall lcd_impulsR          ;7 Takte warten
; in wl,lcdport-2            ;im Readmode LCD auslesen
 cbi enable1port,enable1    ;Enable wieder L
 cbi enable2port,enable2    ;Enable wieder L
lcd_impulsR: 
 ret                        ;und wieder zurck...


;============================================================================
; Interne Routine, erzeugt (vernichtet??) Wartezeit
;----------------------------------------------------------------------------
delay5ms:               ;5ms Pause fr manche Befehle
 push zl                    ;alten Inhalt
 push zh                    ;sichern
 ldi zh,high(clock/11000*5) ;Verzgerungswert fr 5ms
 ldi zl,low(clock/11000*5)  ;setzen
delay5ms1:
 rcall delay5ms2            ;7 Takte vertrdeln
 sbiw zh:zl,1               ;-1 (2 weitere Takte)
 brne delay5ms1             ;0? nein, nochmal (2 Takte)
 pop zh                     ;alten Inhalt
 pop zl                     ;zurck holen
delay5ms2: 
 ret                        ;zurck...

